/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* links.c: manage links between people */

#include "structs.h"

transitive(prs2, prs3,
relcode, othercode, relation,
mardate, adopt, adoptdate)
long prs2, prs3;
char relcode, othercode, *relation;
date mardate, adoptdate;
{
char parentflag;
date cbd, cdd;
struct person *p;
int i;
char fname2[NAMELEN], fname3[NAMELEN];
struct prslink il;

p = select_person(prs2);
cbd = p->birth;
cdd = p->death;

/* if married after the child is an adult, no implied link */
if(cbd && (mardate - cbd) / YEAR_DAY_RATIO > 18) return;

/* if marriage after child's death, no implied link */
if(cdd && mardate > cdd) return;

/* set up for third link */
init_0(&il);
il.prs1 = prs3;
il.prs2 = prs2;
il.linktype = relcode;

birthname(prs2);
strcpy(fname2, fld1);
birthname(prs3);
strcpy(fname3, fld1);

/* see if perhaps prs3 is already a parent (adopting) of prs2 */
spreadout(prs2, relcode);
parentflag = 0;
for(i=0; i<spreadnum; ++i) {
if(spreaddata[i].prsid == prs3) return;
if(!spreaddata[i].adopt) parentflag = 1;
}

/* maybe this spouse is the parent */
/* a meaningless question to ask if the child was adopted */
if(!adopt && !parentflag) {
if(!cbd || (mardate-cbd) / YEAR_DAY_RATIO < 3) {
printf("is %s the %s of %s?",
fname3, relation, fname2);
if(get_yn("")) {
insupd_prslink(NULLKEY, &il);
return;
}
}
}

/* check for custidy */
spreadout(prs2, othercode);
if(spreadnum > 1) return;

printf("is %s a caregiving %s of %s?",
fname3, relation, fname2);
if(!get_yn("")) return;
if(adopt && adoptdate > cbd) cbd = adoptdate;
if(mardate > cbd) cbd = mardate;
il.end = cbd;
il.adopt = 1;
insupd_prslink(NULLKEY, &il);

if(relcode == 'f')
inherit_lname(prs3, prs2, cbd);
} /* transitive */

static struct SPREADDATA kids[SPREADMAX];
static int nkids;

single_parent(prs1, prs2, mardate, relcode, othercode, relation)
long prs1, prs2;
date mardate;
char relcode, othercode;
char *relation;
{
int i;
date adoptdate;

spreadout(prs1, 'c');
memcpy(kids, spreaddata, sizeof(kids));
nkids = spreadnum;

for(i=0; i<nkids; ++i) {
adoptdate = 0;
if(kids[i].adopt)
adoptdate = (select_prslink(kids[i].rownum))->end;
transitive(kids[i].prsid, prs2, relcode, othercode, relation,
mardate, kids[i].adopt, adoptdate);
} /* end loop over kids */
} /* single_parent */

/* make the spouse a parent */
spouseparent(prs1, prs2, spousetype, relcode, othercode, relation,
adopt, adoptdate)
long prs1, prs2;
char spousetype, relcode, othercode, *relation;
date adoptdate;
{
long prs3;
date mardate; /* marriage date */

spreadout(prs1, spousetype);
if(!spreadnum) return;
--spreadnum;
prs3 = spreaddata[spreadnum].prsid;
mardate = spreaddata[spreadnum].order;

transitive(prs2, prs3, relcode, othercode, relation,
mardate, adopt, adoptdate);
} /* spouseparent */

/* add any implied links */
implied_links(l)
struct prslink *l;
{
char linktype = l->linktype;

/* check for name change */
if(linktype == 'h' || l->adopt && linktype == 'f')
inherit_lname(l->prs1, l->prs2, l->adopt ? l->end : l->start);

if(linktype == 'h') {
/* find any unclaimed children */
single_parent(l->prs1, l->prs2, l->start, 'm', 'f', "mother");
single_parent(l->prs2, l->prs1, l->start, 'f', 'm', "father");
return;
} /* end spouse link */

/* new child added, is most recent spouse the parent or guardian? */
if(linktype == 'm')
spouseparent(l->prs1, l->prs2, 'h', 'f', 'm', "father", l->adopt, l->end);
else
spouseparent(l->prs1, l->prs2, 'w', 'm', 'f', "mother", l->adopt, l->end);
} /* implied_links */

/* spread out all links for a person */
spreadout(prsid, linktype)
long prsid;
char linktype; /* should be one of fmchw */
{
struct rel_cursor c;
date when;
char *keyname;
struct prslink l, *l1;

spreadnum = 0;
if(strchr("hmf", linktype))
keyname = "prs2";
else
keyname = "prs1";
if(linktype == 'w') linktype = 'h';

IDX_LOOP_prslink(&c, l1, &l, prsid, keyname)
if(linktype != l.linktype) {
if(linktype != 'c') continue;
if(l.linktype == 'h') continue;
}

/* we have a valid link to this person */
if(spreadnum == SPREADMAX)
eprint("too many links for %ld, internal error", prsid);

when = l.start;
if(linktype == 'c')
when = (select_person(l.prs2))->birth;
if(linktype == 'm' || linktype == 'f') {
/* put the birth mother or father first */
if(!l.adopt) when = EARLY_DATE;
else when = l.end;
}

spread_insert(c.rowid, l.prs1+l.prs2 - prsid, when, l.adopt);
END_LOOP_prslink(&c)
} /* spreadout */

spread_insert(rownum, prsid, order, adopt)
long rownum, prsid, order;
char adopt;
{
int i, j;
struct SPREADDATA *s;

if(order) { /* a valid order */
s = spreaddata;
for(i=0; i<spreadnum; ++i,++s) {
if(!s->order) break;
if(s->order > order) break;
}

/* slide the remaining down and put this one in position */
j = spreadnum-1;
if(spreadnum == SPREADMAX) --j;
else ++spreadnum;
s = spreaddata + j;
for(; j>=i; --j,--s)
s[1] = s[0]; /* structure copy */
if(i < SPREADMAX) {
s = spreaddata + i;
s->order = order;
s->prsid = prsid;
s->rownum = rownum;
s->adopt = adopt;
}
} else {

/* no date present, just stick it on the end */
if(spreadnum < SPREADMAX) {
s = spreaddata + spreadnum;
s->order = order;
s->prsid = prsid;
s->rownum = rownum;
s->adopt = adopt;
++spreadnum;
}
}
} /* spread_insert */

long getmom(prsid)
long prsid;
{
spreadout(prsid, 'm');
if(spreadnum && !spreaddata[0].adopt)
return spreaddata[0].prsid;
return 0L;
} /* getmom */

long getdad(prsid)
long prsid;
{
spreadout(prsid, 'f');
if(spreadnum && !spreaddata[0].adopt)
return spreaddata[0].prsid;
return 0L;
} /* getdad */

/* scan through links, looking to update the dates */
linkscan_date(linktype)
char linktype;
{
long spouse;
struct prslink new_l;
int i, change;
date save_aniv, d;

spreadout(cur_prs_rowid, linktype);

for(i=0; i<spreadnum; ++i) {
change = 0;
select_copy_prslink(spreaddata[i].rownum, &new_l);
spouse = spreaddata[i].prsid;
birthname(spouse);
if(new_l.start)
printf("Marriage with %s begins on %s\n", fld1, datestring(new_l.start));
else
printf("Marriage with %s has no wedding date\n", fld1);
lifetime_dates(&cur_prs, cur_fname, 10, 120);
save_aniv = new_l.start;
d = getdate("wedding date", datemsg, 0);
if(d) { new_l.start = d; change = 1; }
if(new_l.start) {
birthname(spouse);
if(new_l.end)
printf("Marriage with %s ends on %s\n", fld1, datestring(new_l.end));
lowdate = new_l.start;
highdate = lowdate + 100*YEAR_DAY_RATIO;
if(cur_prs.death) highdate = cur_prs.death;
d = getdate("date of divorce", "date does not lie within the span of the couple's marriage", 0);
if(d) { new_l.end = d; change = 1; }
}

if(change) {
insupd_prslink(spreaddata[i].rownum, &new_l);
up_aniv(new_l.prs1, new_l.prs2, save_aniv, new_l.start);
up_aniv(new_l.prs2, new_l.prs1, save_aniv, new_l.start);
}
} /* end looking over spouses */
} /* linkscan_date */

up_aniv(prsid, spouse, aniv1, aniv2)
long prsid, spouse;
date aniv1, aniv2;
{
struct rel_cursor c;
struct name n, *n1;
char prstype;
struct prslink l, *l1;
int j;
int cnt_lname;
long save_rowid;
char lname[NAMELEN];
long list[SPREADMAX];

prstype = (select_person(prsid))->sex;
if(prstype == 'f') {
birthname(spouse);
strcpy(lname, fld3);
}

save_rowid = 0;
cnt_lname = 0;
j = 0;
IDX_LOOP_name(&c, n1, &n, prsid, "prsid")
if(n.start == aniv1)
list[j++] = c.rowid;
if(prstype == 'm' || n.birth || n.start) continue;
splitstring(n.fullname);
if(strcmp(fld3, lname)) continue;
++cnt_lname;
save_rowid = c.rowid;
END_LOOP_name(&c)

if(cnt_lname == 1) list[j++] = save_rowid;

while(--j >= 0) {
select_copy_name(list[j], &n);
n.start = aniv2;
insupd_name(list[j], &n);
}

j = 0;
IDX_LOOP_prslink(&c, l1, &l, prsid, "prs1")
if(l.adopt && l.end == aniv1)
list[j++] = c.rowid;
END_LOOP_prslink(&c)

while(--j >= 0) {
select_copy_prslink(list[j], &l);
l.end = aniv2;
insupd_prslink(list[j], &l);
}
} /* up_aniv */

inherit_lname(prs1, prs2, startdate)
long prs1, prs2;
date startdate;
{
struct rel_cursor c;
struct name n, *n1;
int namefound;
char lname[NAMELEN]; /* husband/father last name */

birthname(prs1);
strcpy(lname, fld3);

namefound = 0;
IDX_LOOP_name(&c, n1, &n, prs2, "prsid")
splitstring(n.fullname);
if(!strcmp(fld3, lname)) { namefound = 1; break; }
END_LOOP_name(&c)

if(namefound) return;

birthname(prs2);
printf("did %s take the name %s?", fld1, lname);
if(!get_yn("")) return;

select_copy_name(birth_rowid, &n);
n.start = startdate;
change_lname(&n, lname);
n.birth = 0;
insupd_name(NULLKEY, &n);
} /* inherit_lname */

up_gender()
{
char c = cur_prs.sex;
int i;
struct prslink new_l;

spreadout(cur_prs_rowid, (c == 'm') ? 'w' : 'h');
if(spreadnum) {
printf("%s is already married and cannot switch gender\n", cur_fname);
return;
}

spreadout(cur_prs_rowid, 'c');
nkids = 0;
for(i=0; i<spreadnum; ++i)
if(!spreaddata[i].adopt)
kids[nkids++].prsid = spreaddata[i].prsid;

for(i=0; i<nkids; ++i) {
spreadout(kids[i], c);
if(spreadnum && !spreaddata[0].adopt) {
birthname(kids[i]);
printf("%s has a child %s who already has a %s, cannot switch gender\n",
cur_fname, fld1, c == 'f' ? "father" : "mother");
return;
}
} /* loop over children */

/* ok to make the switch */
spreadout(cur_prs_rowid, 'c');
for(i=0; i<spreadnum; ++i) {
select_copy_prslink(spreaddata[i].rownum, &new_l);
new_l.linktype = c;
insupd_prslink(spreaddata[i].rownum, &new_l);
}

c = 'm'+'f' - c;
cur_prs.sex = c;
insupd_person(cur_prs_rowid, &cur_prs);
} /* up_gender */

/* drop names and shortnotes, and then the person */
delperson(prsid)
long prsid;
{
casc_upd_del(prsid, 0L, STRUCT_address, "prsid");
casc_upd_del(prsid, 0L, STRUCT_qknote, "prsid");
casc_upd_del(prsid, 0L, STRUCT_name, "prsid");
delete_person(prsid);
} /* delperson */

zap_person()
{
int i, firstyear;
char linktype;
char *idxname;
struct prslink *l;
struct rel_cursor c;

spreadout(cur_prs_rowid, 'm');
if(spreadnum) goto linksfound;
spreadout(cur_prs_rowid, 'f');
if(spreadnum) goto linksfound;
spreadout(cur_prs_rowid, 'h');
if(spreadnum) goto linksfound;
spreadout(cur_prs_rowid, 'w');
if(spreadnum) goto linksfound;
spreadout(cur_prs_rowid, 'c');

linksfound:
if(!spreadnum) {
printf("\n\nThis command will remove the person %s %s from the database\n", cur_fname, cur_lname);
if(!get_yn("Do you really want to do this?")) return;

textstats(cur_prs_rowid, &i, &firstyear);
if(i) {
printf("%s has %d text entries, please remove these text entries first\n", cur_fname, i);
return;
}

delperson(cur_prs_rowid);
cur_prs_rowid = 0L;
printf("Person deleted\n\n");
longjmp(main_menu, 1);
}

printf("\n\nThis command deletes links associated with %s %s.\n", cur_fname, cur_lname);
printf("Which links shall I remove, (father mother %s children all)? ",
cur_prs.sex == 'm' ? "wife" : "husband");
linktype = getletter("hwmfca");
nl();

idxname = "prs2";
if(linktype == 'w' || linktype == 'c')
idxname = "prs1";
if(linktype == 'w') linktype = 'h';

IDX_LOOP_prslink(&c, l, NULL, cur_prs_rowid, idxname)
if(linktype != 'a') {
if(l->linktype != linktype) {
if(linktype != 'c') continue;
if(l->linktype == 'h') continue;
}
}
delete_prslink(c.rowid);
END_LOOP_prslink(&c)

if(linktype == 'a')
casc_upd_del(cur_prs_rowid, NULLKEY, STRUCT_prslink, "prs1");
} /* zap_person */

